;; Generated package description from tramp.el  -*- no-byte-compile: t -*-
(define-package "tramp" "2.7.2.4" "Transparent Remote Access, Multiple Protocol" '((emacs "27.1")) :commit "c37e5fa5f4e60a19b38089672fbad377035a33a2" :authors '(("Kai Großjohann" . "kai.grossjohann@gmx.net") ("Michael Albinus" . "michael.albinus@gmx.de")) :maintainer '("Michael Albinus" . "michael.albinus@gmx.de") :keywords '("comm" "processes") :url "https://www.gnu.org/software/tramp/")
